import json
import logging
import sys
import time
import traceback

from redis import StrictRedis

from minique.enums import JobStatus
from minique.excs import AlreadyAcquired, AlreadyResulted
from minique.models.job import Job
from minique.utils import _set_current_job, import_by_string


class JobRunner:
    def __init__(self, worker, job):
        assert isinstance(job, Job)
        self.worker = worker
        self.job = job
        self.redis = job.redis
        assert isinstance(self.redis, StrictRedis)
        self.log = logging.getLogger('{}.{}'.format(__name__, self.job.id.replace('.', '_')))
        job.ensure_exists()

    def acquire(self):
        new_acquisition_info = json.dumps({'worker': self.worker.id, 'time': time.time()})
        if not self.redis.hsetnx(self.job.redis_key, 'acquired', new_acquisition_info):
            raise AlreadyAcquired('job {id} already acquired: {info}'.format(
                id=self.job.id,
                info=self.job.acquisition_info,
            ))
        self.redis.hset(self.job.redis_key, 'status', JobStatus.ACQUIRED.value)
        self.redis.persist(self.job.redis_key)

    def execute(self):
        func = self.get_callable()
        kwargs = self.job.kwargs
        self.log.info('calling %s(%r)', func, kwargs)
        with _set_current_job(job=self.job):
            return func(**kwargs)

    def get_callable(self):
        return import_by_string(self.job.callable_name)

    def complete(self, success, value, duration):
        assert isinstance(success, bool)
        update_payload = {
            'status': (JobStatus.SUCCESS if success else JobStatus.FAILED).value,
            'duration': float(duration),
        }
        if not self.redis.setnx(self.job.result_redis_key, value):  # pragma: no cover
            raise AlreadyResulted('job {id} already has result'.format(id=self.job.id))
        self.redis.hmset(self.job.redis_key, update_payload)
        # Update expiries to the result TTL for both the job and the result
        self.redis.expire(self.job.result_redis_key, self.job.result_ttl)
        self.redis.expire(self.job.redis_key, self.job.result_ttl)
        if success:
            self.log.info('finished in %f seconds', duration)
        else:
            self.log.warning('errored in %f seconds', duration)

    def run(self):
        self.acquire()
        interrupt = False
        success = False
        value = {'error': 'unknown'}
        start_time = time.time()
        try:
            value = self.execute()
            value = json.dumps(value)
            success = True
        except BaseException as exc:
            success = False
            exc_type, exc_value, exc_tb = excinfo = sys.exc_info()
            value = json.dumps({
                'exception_type': exc_type.__qualname__,
                'exception_value': str(exc_value),
                'traceback': traceback.format_exc(),
            }, default=str)
            interrupt = isinstance(exc, KeyboardInterrupt)
            try:
                self.process_exception(excinfo)
            except:  # noqa
                self.log.warning('error running process_exception()', exc_info=True)
        finally:
            end_time = time.time()
            self.complete(success=success, value=value, duration=(end_time - start_time))
        if interrupt:  # pragma: no cover
            raise KeyboardInterrupt('Interrupt')

    def process_exception(self, excinfo):  # pragma: no cover
        """
        A hook for subclasses to log (e.g. Sentry) or otherwise process exceptions.

        :param excinfo: The sys.exc_info() 3-tuple
        """
        pass



import json
import time

from minique.enums import JobStatus
from minique.excs import DuplicateJob
from minique.models.job import Job
from minique.models.queue import Queue
from minique.utils import get_random_pronounceable_string


def enqueue(redis, queue_name, callable, kwargs=None, job_id=None, job_ttl=0, result_ttl=86400 * 7):
    if not isinstance(callable, str):
        callable = '{module}.{qualname}'.format(
            module=callable.__module__,
            qualname=callable.__qualname__,
        )

    if not job_id:
        job_id = get_random_pronounceable_string()
    job = Job(redis, job_id)
    if job.exists:
        raise DuplicateJob('duplicate job: {id}'.format(id=job_id))
    payload = {
        'queue': queue_name,
        'callable': str(callable),
        'kwargs': json.dumps(kwargs or {}),
        'status': JobStatus.NONE.value,
        'ctime': time.time(),
        'job_ttl': int(job_ttl),
        'result_ttl': int(result_ttl),
    }
    queue = Queue(redis, name=queue_name)
    with redis.pipeline() as p:
        p.hmset(job.redis_key, payload)
        if payload['job_ttl'] > 0:
            p.expire(job.redis_key, payload['job_ttl'])
        p.rpush(queue.redis_key, job.id)
        p.execute()
    job.ensure_exists()
    return job


def get_job(redis, job_id):
    job = Job(redis, job_id)
    job.ensure_exists()
    return job


def cancel_job(redis, job_id):
    """
    Cancel the job with the given job ID.

    If a worker is already busy with the job, it may not immediately quit,
    and as such, the job is not set to cancelled state.

    :type redis: redis.StrictRedis
    :param job_id: Job ID.
    :raises minique.excs.NoSuchJob: if the job does not exist.
    """
    job = get_job(redis, job_id)
    if not (job.has_finished or job.has_started):
        # Cancel the job and remove it from the queue it may be in.
        redis.hset(job.redis_key, 'status', JobStatus.CANCELLED.value)
        redis.lrem(job.get_queue().redis_key, 0, job.id)
        return True
    return False



from minique.utils import _set_current_job, import_by_string
from minique.work.job_runner import JobRunner
from minique.work.worker import Worker


class TestJobRunner(JobRunner):
    def execute(self):
        func = (self.job.replacement_callable or import_by_string(self.job.callable_name))
        if self.job.replacement_kwargs is not None:
            kwargs = self.job.replacement_kwargs
        else:
            kwargs = self.job.kwargs
        with _set_current_job(job=self.job):
            return func(**kwargs)


def run_synchronously(job, replacement_callable=None, replacement_kwargs=None):
    job.replacement_callable = replacement_callable
    job.replacement_kwargs = replacement_kwargs
    worker = Worker(job.redis, [])
    TestJobRunner(worker=worker, job=job).run()
